unit fMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  FlexBase, FlexUtils, FlexControls, FlexProps, Utils, ComCtrls, StdCtrls,
  ImgList, ExtCtrls, ToolWin, Buttons, ACCtrls, FlexPath;

type
  TfmMain = class(TForm)
    sbMain: TStatusBar;
    tbMain: TTabControl;
    tmSelected: TTimer;
    Panel1: TPanel;
    Panel2: TPanel;
    sbtControl: TSpeedButton;
    sbtZoomIn: TSpeedButton;
    sbtZoomOut: TSpeedButton;
    sbtOptions: TSpeedButton;
    sbtAbout: TSpeedButton;
    sbtExit: TSpeedButton;
    tmLamps: TTimer;
    tmTimeOuts: TTimer;
    fpMain: TFlexPanel;
    procedure FormCreate(Sender: TObject);
    procedure tbMainChange(Sender: TObject);
    procedure fpMainMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure fpMainNotify(Sender: TObject; Control: TFlexControl;
      Notify: TFlexNotify);
    procedure FormDestroy(Sender: TObject);
    procedure sbtExitClick(Sender: TObject);
    procedure sbtControlClick(Sender: TObject);
    procedure fpMainExit(Sender: TObject);
    procedure sbtOptionsClick(Sender: TObject);
    procedure fpMainPaintOver(ACanvas: TCanvas; AControl: TFlexControl;
      ChildrenOnly, SelectedOnly: Boolean);
    procedure tmSelectedTimer(Sender: TObject);
    procedure fpMainMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure sbtZoomClick(Sender: TObject);
    procedure sbtAboutClick(Sender: TObject);
    procedure tmLampsTimer(Sender: TObject);
    procedure tmTimeOutsTimer(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    { Private declarations }
    FCopyright: TFlexText;
    FSelectionBrush: TBrush;
    FSelected: TFlexControl;
    FSelectedDisp: integer;
    FZoneNamesLayer: TFlexLayer;
    FZoneNames: TList;
    FZones: TList;
    FDoors: TList;
    FLamps: TList;
    FLampsDisp: integer;
    FLastActive: TFlexControl;
    FLastActiveStyle: TColorStyles;
    FSchemeBtn: TFlexControl;
    FDoorsTimeOuts: TList;
    FMinimized: boolean;
    procedure SortLamps;
    function  GetSchemeBtn(PTabIndex: PInteger = Nil): TFlexControl;
    procedure PaintActive(AsActive: boolean);
    procedure PaintSchemeBtn(AsActive: boolean);
    procedure SetSelected(const Value: TFlexControl);
    function  CreateSelectedRegion(ForInvalidate: boolean = false): HRGN;
    procedure RefreshSelected;
    procedure EventsChange(Sender: TObject);
    procedure SetDoorTimeOut(Door: TFlexDoor; PassId: integer);
    procedure AppMinimize(Sender: TObject);
    procedure AppMaximize(Sender: TObject);
    procedure PaintMarquee(DC: HDC);
  public
    { Public declarations }
    procedure SetOptions;
    function  FindZone(ZoneId: integer): TFlexControl;
    function  FindPass(PassId: integer): TFlexControl;
    function  DoPass(PassId, PersonId: integer; DoOpen: boolean): boolean;
    procedure DoBlock(PassId: integer; IsBlock: boolean);
    property  Selected: TFlexControl read FSelected write SetSelected;
  end;

var
  fmMain: TfmMain;

implementation

{$R *.DFM}

uses
  dMain, fControl, fOptions, fAboutPrg, fSplash, fPersCard, FlexCards;

type
  PDoorTimeOut = ^TDoorTimeOut;
  TDoorTimeOut = record
   Door: TFlexDoor;
   PassId: integer;
   TimeToClose: TDateTime;
  end;

procedure TfmMain.FormCreate(Sender: TObject);
var i, NewWidth, NewHeight: integer;
begin
 // Creating lists
 FZoneNames := TList.Create;
 FZones := TList.Create;
 FDoors := TList.Create;
 FLamps := TList.Create;
 FDoorsTimeOuts := TList.Create;
 // Creating selection brush
 FSelectionBrush := CreateSelectionBrush(clNavy);
 // Loading from file
 LoadFlexCursors;
 fpMain.OnProgress := fmSplash.FlexProgress;
 Screen.Cursor := crHourGlass;
 try
  fpMain.LoadFromFile(ExtractFilePath(ParamStr(0))+'access.fxd');
 finally
  fpMain.OnProgress := Nil;
  Screen.Cursor := crDefault;
 end;
 // Cache all schemes bitmap
 for i:=0 to fpMain.Schemes.Count-1 do
  (fpMain.Schemes[i] as TFlexScheme).BrushProp.BitmapCache := true;
 // Sort lamp controls
 SortLamps;
 // Adjust window size
 fpMain.HorzScrollBar.Visible := False;
 fpMain.VertScrollBar.Visible := False;
 NewWidth := Width + (fpMain.DocWidth - fpMain.ClientWidth);
 NewHeight := Height + (fpMain.DocHeight - fpMain.ClientHeight);
 if (NewWidth > Screen.Width) or (NewHeight > Screen.Height) then
  WindowState := wsMaximized
 else begin
  Width := NewWidth;
  Height := NewHeight;
 end;
 fpMain.HorzScrollBar.Visible := True;
 fpMain.VertScrollBar.Visible := True;
 // Making layer "ZoneNames" non selectable
 if Assigned(FZoneNamesLayer) then FZoneNamesLayer.Selectable := false;
 fpMain.Layers.ByName['Building'].Selectable := false;
 FCopyright.Hint := 'About';
 // Load and set options from ini-file
 LoadOptions;
 SetOptions;
 // Start event
 dmMain.OnEventAdd := EventsChange;
 dmMain.AddEvent('System online', 0, 0, 0);
 tmTimeOuts.Enabled := true;
 Application.OnMinimize := AppMinimize;
 Application.OnRestore := AppMaximize;
end;

procedure TfmMain.FormDestroy(Sender: TObject);
begin
 dmMain.tmMovements.Enabled := false;
 DestroyAllFlexCards;
 HideAllPersonCards;
 // Destroy selection brush
 if Assigned(FSelectionBrush) then begin
  FSelectionBrush.Bitmap.Free;
  FSelectionBrush.Bitmap := Nil;
  FSelectionBrush.Free;
  FSelectionBrush := Nil;
 end;
 // Destroy lists
 FZoneNames.Free;
 FZones.Free;
 FDoors.Free;
 FLamps.Free;
 // Destroy door timeouts
 while FDoorsTimeOuts.Count > 0 do begin
  Dispose(PDoorTimeOut(FDoorsTimeOuts[FDoorsTimeOuts.Count-1]));
  FDoorsTimeOuts.Delete(FDoorsTimeOuts.Count-1);
 end;
 FDoorsTimeOuts.Free;
end;

procedure TfmMain.AppMinimize(Sender: TObject);
begin
 FMinimized := true;
 HideAllPersonCards;
end;

procedure TfmMain.AppMaximize(Sender: TObject);
begin
 FMinimized := false;
end;

procedure TfmMain.SetSelected(const Value: TFlexControl);
var Index: integer;
    Rgn: HRGN;
begin
 if Value = FSelected then exit;
 // Invalidate frame around old selected control
 if Assigned(FSelected) then begin
  //RefreshSelected;
  Rgn := CreateSelectedRegion(True);
  InvalidateRgn(fpMain.Handle, Rgn, false);
  DeleteObject(Rgn);
  FSelected := Nil;
 end;
 // Check Value parent scheme
 if Assigned(Value) and (Value.ParentScheme <> fpMain.ActiveScheme) then
  fpMain.ActiveScheme := Value.ParentScheme;
 // Set new selected control
 FSelected:= Value;
 if Assigned(FSelected) then begin
  // Start refresh timer
  tmSelected.Enabled := true;
  // Invalidate frame around new selected control
  RefreshSelected;
  if Assigned(fmControl) then begin
   // Locate object in lists
   if FSelected is TFlexDoor then begin
    // Locate PassA of FlexDoor
    Index := dmMain.FindPass(TFlexDoor(FSelected).PassAProp.Value);
    if Index >= 0 then begin
     fmControl.lbPasses.ItemIndex := Index;
     fmControl.pgTables.ActivePage := fmControl.tsPasses;
    end else
     fmControl.lbPasses.ItemIndex := -1;
    fmControl.RefreshPassInfo;
   end else
   if TFlexControl(FSelected).UserData.Values['ZoneID'] <> '' then begin
    // Locate zone
    Index := StrToIntDef(TFlexControl(FSelected).UserData.Values['ZoneID'], 0);
    Index := dmMain.FindZone(Index);
    if Index >= 0 then begin
     fmControl.lbZones.ItemIndex := Index;
     fmControl.pgTables.ActivePage := fmControl.tsZones;
    end else
     fmControl.lbZones.ItemIndex := -1;
    fmControl.RefreshZoneInfo;
   end;
  end;
 end;
end;

function TfmMain.CreateSelectedRegion(ForInvalidate: boolean = false): HRGN;
var GroupRect: TRect;
    GroupInited: boolean;

 procedure ProcessControl(Control: TFlexControl);
 var i: integer;
     Points: TPointArray;
     Rgn: HRGN;
 begin
  Points := Nil;
  if Control is TFlexGroup then begin
   // Process group controls
   for i:=0 to Control.Count-1 do ProcessControl(Control[i]);
  end else
  if ForInvalidate then begin
   // Create region for bounding box
   if GroupInited then
    UnionRect(GroupRect, GroupRect, Control.PaintRect)
   else begin
    GroupRect := Control.PaintRect;
    GroupInited := true;
   end;
  end else begin
   if Control.PointCount > 0 then with Control.PaintRect do begin
    // Process polyline
    Points := Control.GetTransformPoints(Left, Top, fpMain.Scale);
    Rgn := CreatePolygonRgn(Points[0], Length(Points), ALTERNATE);
   end else
   if Control is TFlexEllipse then
    // Process ellipse
    Rgn := CreateEllipticRgnIndirect(Control.PaintRect)
   else
   if Control is TFlexBox then with Control.PaintRect do begin
    // Process rectangle
    if TFlexBox(Control).RoundnessProp.Value > 0 then begin
     i := ScaleValue(TFlexBox(Control).RoundnessProp.Value, Control.Owner.Scale);
     Rgn := CreateRoundRectRgn(Left, Top, Right+1, Bottom+1, i, i);
    end else
     Rgn := CreateRectRgnIndirect(Control.PaintRect);
   end else
    // Process unknown other control (as rect)
    Rgn := CreateRectRgnIndirect(Control.PaintRect);
   if Rgn = 0 then exit;
   if Result = 0 then
    Result := Rgn
   else begin
    CombineRgn(Result, Result, Rgn, RGN_OR);
    DeleteObject(Rgn);
   end;
  end;
 end;

begin
 Result := 0;
 if not Assigned(FSelected) then exit;
 GroupInited := not ForInvalidate;
 ProcessControl(FSelected);
 if ForInvalidate and GroupInited then with GroupRect do
  Result := CreateRectRgn(Left, Top, Right+1, Bottom+1);
end;

procedure TfmMain.PaintMarquee(DC: HDC);
var Rgn: HRGN;
    P: TPoint;
begin
 // Paint selected frame
 Rgn := CreateSelectedRegion;
 if Rgn = 0 then exit;
 SetBrushOrgEx(DC, 7-FSelectedDisp, 7-FSelectedDisp, @P);
 FrameRgn(DC, Rgn, FSelectionBrush.Handle,
   Options.SelFrameWidth, Options.SelFrameWidth);
 SetBrushOrgEx(DC, P.X, P.Y, Nil);
 DeleteObject(Rgn);
end;

procedure TfmMain.fpMainPaintOver(ACanvas: TCanvas; AControl: TFlexControl;
  ChildrenOnly, SelectedOnly: Boolean);
begin
 if Assigned(FSelected) and tmSelected.Enabled then
  PaintMarquee(ACanvas.Handle);
end;

procedure TfmMain.RefreshSelected;
var DC: HDC;
begin
 if not Assigned(FSelected) then exit;
 DC := GetDC(fpMain.Handle);
 try
  PaintMarquee(DC);
 finally
  ReleaseDC(fpMain.Handle, DC);
 end;
end;

procedure TfmMain.tmSelectedTimer(Sender: TObject);
begin
 dec(FSelectedDisp);
 FSelectedDisp := FSelectedDisp and 7;
 RefreshSelected;
end;

procedure TfmMain.tmLampsTimer(Sender: TObject);
const
  NumColors = 6;
  RunColors: array[0..NumColors-1] of TColor = (
   // $00800000, $00901D0A, $00A84717, $00BF7125, $00D89B32, $00EEC440 );
      $000000FF, $00002AFF, $000064FF, $00009CFF, $0000D5FF, $0000FFFF );
   // $00800000, $00800000, $00800000, $00800000, $00800000, $00FFFFFF );
var i, Idx: integer;
begin
 if FLampsDisp = 0
  then FLampsDisp := NumColors - 1
  else dec(FLampsDisp);
 for i:=0 to FLamps.Count-1 do with TFlexEllipse(FLamps[i]) do
  case Options.LampsEffect of
   leRunning:
     begin
      Idx := (i+FLampsDisp) mod NumColors;
      BrushProp.Color := RunColors[Idx];
     end;
   leFlashing:
     if (i+FLampsDisp) and 1 = 0
      then BrushProp.Color := clNavy
      else BrushProp.Color := clAqua;
  end;
end;

function TfmMain.FindZone(ZoneId: integer): TFlexControl;
var i: integer;
begin
 Result := Nil;
 for i:=0 to FZones.Count-1 do
  if StrToIntDef(TFlexControl(FZones[i]).UserData.Values['ZoneID'], 0) =
     ZoneId then begin
   Result := TFlexControl(FZones[i]);
   break;
  end;
end;

function TfmMain.FindPass(PassId: integer): TFlexControl;
var i: integer;
begin
 Result := Nil;
 for i:=0 to FDoors.Count-1 do
  if (TFlexDoor(FDoors[i]).PassAProp.Value = PassId) or
     (TFlexDoor(FDoors[i]).PassBProp.Value = PassId) then begin
   Result := TFlexControl(FDoors[i]);
   break;
  end;
end;

procedure TfmMain.tbMainChange(Sender: TObject);
var Scheme: TFlexControl;
begin
 Scheme := Nil;
 case tbMain.TabIndex of
  0: Scheme := fpMain.FindControl('MainPlan');
  1: Scheme := fpMain.FindControl('ControlPost');
  2: Scheme := fpMain.FindControl('AdmGrandFloor');
  3: Scheme := fpMain.FindControl('ProdGrandFloor');
  4: Scheme := fpMain.FindControl('TechGrandFloor');
 end;
 if Assigned(Scheme) and (fpMain.ActiveScheme <> TFlexScheme(Scheme)) then begin
  HideAllPersonCards;
  fpMain.ActiveScheme := TFlexScheme(Scheme);
 end;
end;

procedure TfmMain.SetOptions;
var i: integer;
    Layer: TFlexLayer;
begin
 fpMain.ShowHint := Options.ShowHints;
 case Options.ShowZoneCaptions of
  zsNone:
    begin
     // Make ZoneNames layer invisible
     if Assigned(FZoneNamesLayer) then FZoneNamesLayer.Visible := False;
    end;
  zsSelected:
    begin
     // Hide all ZoneName controls
     for i:=0 to FZoneNames.Count-1 do
      TFlexControl(FZoneNames[i]).Visible := False;
     // Make ZoneNames layer visible
     if Assigned(FZoneNamesLayer) then FZoneNamesLayer.Visible := True;
    end;
  zsAll:
    begin
     // Show all ZoneName controls
     for i:=0 to FZoneNames.Count-1 do
      TFlexControl(FZoneNames[i]).Visible := true;
     // Make ZoneNames layer visible
     if Assigned(FZoneNamesLayer) then FZoneNamesLayer.Visible := True;
    end;
 end;
 Layer := fpMain.Layers.ByName['Lamps'];
 case Options.LampsEffect of
  leNone:
    begin
     if Assigned(Layer) then Layer.Visible := false;
    end;
  leFlashing,
  leRunning:
    begin
     if Assigned(Layer) then Layer.Visible := true;
     tmLamps.Enabled := true;
    end;
 end;

 if not Options.ShowPersonsCards or not Options.Emulation then begin
  HideAllPersonCards;
  DestroyAllFlexCards;
 end else
 if Options.ShowCardsInFlex
  then HideAllPersonCards
  else DestroyAllFlexCards;
 // Emualtion
 dmMain.tmMovements.Enabled := Options.Emulation;
end;

procedure TfmMain.SortLamps;
var Sorted: boolean;
    i: integer;
    Lamp: pointer;
begin
 // Bubble sorting
 repeat
  Sorted := true;
  for i:=1 to FLamps.Count-1 do
   if TFlexControl(FLamps[i]).Tag < TFlexControl(FLamps[i-1]).Tag then begin
    Lamp := FLamps[i];
    FLamps[i] := FLamps[i-1];
    FLamps[i-1] := Lamp;
    Sorted := false;
   end;
 until Sorted;
end;

procedure TfmMain.PaintActive(AsActive: boolean);
var Style: TColorStyles;
    i, ZoneId, Index: integer;
begin
 if not Assigned(FLastActive) then begin
  FLastActiveStyle := csNone;
  exit;
 end;
 Index := FLastActive.UserData.IndexOfName('ZoneID');
 if Index >= 0 then begin
  Style := csZone;
  if Options.ShowZoneCaptions = zsSelected then begin
   // Show or hide associated ZoneNames
   ZoneId := StrToIntDef(FLastActive.UserData.ValuesByIndex[Index], 0);
   for i:=0 to FZoneNames.Count-1 do
    if TFlexControl(FZoneNames[i]).Tag = ZoneId then
     TFlexControl(FZoneNames[i]).Visible := AsActive;
  end;
 end else
 if FLastActive = FCopyright then
  Style := csCopyright
 else
 if FLastActive.ParentScheme.Name = 'MainPlan'
  then Style := csBuildings
  else Style := csButtons;
 PaintControl(AsActive, FLastActive, Style);
 FLastActiveStyle := Style;
end;

procedure TfmMain.PaintSchemeBtn(AsActive: boolean);
begin
 PaintControl(AsActive, FSchemeBtn, csSchemeBtn);
end;

function TfmMain.GetSchemeBtn(PTabIndex: PInteger = Nil): TFlexControl;
const
  Names: array[0..9, 0..1] of string = (
    ( 'MainPlan', '' ),
    ( 'AdmGrandFloor', 'BtnGrand' ),
    ( 'AdmFloor1', 'BtnFloor1' ),
    ( 'AdmFloor2', 'BtnFloor2' ),
    ( 'ControlPost', '' ),
    ( 'ProdGrandFloor', 'BtnGrand' ),
    ( 'ProdFloor1', 'BtnFloor1' ),
    ( 'ProdFloor2', 'BtnFloor2' ),
    ( 'TechGrandFloor', 'BtnGrand' ),
    ( 'TechFloor1', 'BtnFloor1' )
  );
var
  i: integer;
  Scheme: TFlexCustomScheme;
begin
 if Assigned(PTabIndex) then PTabIndex^ := -1;
 Result := Nil;
 Scheme := fpMain.ActiveScheme;
 if not Assigned(Scheme) then exit;
 for i:=0 to High(Names) do
  if CompareText(Names[i,0], Scheme.Name) = 0 then begin
   if Assigned(PTabIndex) then
    case i of
     0    : PTabIndex^ := 0;
     1..3 : PTabIndex^ := 2;
     4    : PTabIndex^ := 1;
     5..7 : PTabIndex^ := 3;
     8, 9 : PTabIndex^ := 4;
    end;
   if Names[i, 1] <> '' then Result := Scheme.FindByName(Names[i, 1]);
   break;
  end;
end;

procedure TfmMain.fpMainNotify(Sender: TObject; Control: TFlexControl;
  Notify: TFlexNotify);
var BtnControl: TFlexControl;
    TabIdx: integer;
    Id, Index: integer;
    Card: TFlexCard;
begin
 if csDestroying in ComponentState then exit;
 case Notify of
  fnRect:
    if (fpMain.ToolMode = ftmMoving) and
       (Control.UserData.Values['Type'] = 'Card') then begin
     Card := FindFlexCard(Nil, Control);
     if Assigned(Card) then Card.SavePosition;
    end;
  fnLoaded:
    begin
     // Checking Lamp
     if Control.Name = 'Lamp' then begin
      FLamps.Add(Control);
     end else
     // Checking copyright string
     if Control.Name = 'Copyright' then
      FCopyright := TFlexText(Control)
     else
     // Checking ZoneNames layer
     if (Control is TFlexLayer) and (Control.Name = 'ZoneNames') then
      FZoneNamesLayer := TFlexLayer(Control)
     else
     // Checking door control
     if Control is TFlexDoor then begin
      FDoors.Add(Control);
      TFlexDoor(Control).DoorState := dsClosed;
     end else
     // Checking ZoneName controls
     if Control.Name = 'ZoneName' then begin
      FZoneNames.Add(Control);
      Control.Visible := False;
      TFlexText(Control).WordWrap := true;
      Id := StrToIntDef(Control.UserData.Values['ZoneID'], 0);
      Control.Tag := Id;
      Index := dmMain.FindZone(Id);
      if Index >= 0 then
       TFlexText(Control).TextProp.Text := PZone(dmMain.Zones[Index]).Name;
     end else
     // Checking zone controls
     if (Control.UserData.IndexOfName('ZoneID') >= 0) and
        (Control.UserData.IndexOfName('Level') >= 0) then begin
      FZones.Add(Control);
      Id := StrToIntDef(Control.UserData.Values['ZoneID'], 0);
      Index := dmMain.FindZone(Id);
      if Index >= 0 then
       Control.Hint := PZone(dmMain.Zones[Index]).Name;
     end;
    end;
  fnSchemes:
    begin
     if Assigned(FSelected) then
      if fpMain.ActiveScheme <> FSelected.ParentScheme
       then tmSelected.Enabled := false
       else tmSelected.Enabled := true;
     PaintActive(False);
     FLastActive := Nil;
     // Scheme button checking
     BtnControl := GetSchemeBtn(@TabIdx);
     if BtnControl <> FSchemeBtn then begin
      PaintSchemeBtn(False);
      FSchemeBtn := BtnControl;
      PaintSchemeBtn(True);
     end;
     tbMain.TabIndex := TabIdx;
     HideAllPersonCards;
    end;
  fnScale:
    begin
     sbtZoomIn.Enabled := fpMain.Scale < 400;
     sbtZoomOut.Enabled := fpMain.Scale > 50;
    end;
 end;

end;

procedure TfmMain.fpMainMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var Id, Index: integer;
    s: string;

 function IsActive(Control: TFlexControl): boolean;
 begin
  Result := Assigned(Control);
  if not Result then exit;
  Result :=  Assigned(Control.Reference) or
             (Control.UserData.IndexOfName('ZoneID') >= 0) or
             (Control = FCopyright);
 end;

begin
 if not Assigned(fpMain.ActiveScheme) then exit;
 // Checking last active control (button or building on main scheme)
 if fpMain.MouseControl <> FLastActive then begin
  PaintActive(False);
  if IsActive(fpMain.MouseControl) then begin
   FLastActive := fpMain.MouseControl;
   PaintActive(True);
  end else
   FLastActive := Nil;
 end;
 // Update status bar
 s := '';
 if Assigned(FLastActive) then
  case FLastActiveStyle of
   csBuildings:
     begin
      if CompareText(FLastActive.Name, 'Administrate') = 0 then
       s := 'Go to Administrative building plans'
      else
      if CompareText(FLastActive.Name, 'Production') = 0 then
       s := 'Go to Production building plans'
      else
      if CompareText(FLastActive.Name, 'Tech center') = 0 then
       s := 'Go to Technical center plans'
      else
      if CompareText(FLastActive.Name, 'Control Post') = 0 then
       s := 'Go to Control post plan';
      if s <> '' then
       FLastActive.Hint := s;
     end;
   csZone:
     begin
      Id := StrToIntDef(FLastActive.UserData.Values['ZoneID'], 0);
      Index := dmMain.FindZone(id);
      if Index >= 0 then with PZone(dmMain.Zones[Index])^ do
       s := Format('Zone %d: %s', [Id, Name])
      else
       s := 'Unknown zone';
     end;
  end;
 sbMain.SimpleText := s;
end;

procedure TfmMain.fpMainMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 if Sender = fpMain then begin
  fpMain.SetFocus;
  if not fpMain.Focused then Windows.SetFocus(fpMain.Handle);
 end;
 if fpMain.ToolMode <> ftmSelect then exit;
 if (Button = mbLeft) and Assigned(fpMain.MouseSubControl) then begin
  if (FZones.IndexOf(fpMain.MouseSubControl) >= 0) or
     (FDoors.IndexOf(fpMain.MouseSubControl) >= 0) then
   Selected := fpMain.MouseSubControl
  else
  if fpMain.MouseSubControl = FCopyright then begin
   PaintActive(False);
   FLastActive := Nil;
   sbtAbout.Click
  end else
  if fpMain.MouseControl.UserData.Values['Type'] = 'Card' then begin
   fpMain.UnselectAll;
   fpMain.Select(fpMain.MouseControl);
   fpMain.ToolMode := ftmMoving;
  end else
  if not Assigned(fpMain.MouseControl.Reference) then
   Selected := Nil;
 end else
  Selected := Nil;
end;

procedure TfmMain.sbtExitClick(Sender: TObject);
begin
 Close;
end;

procedure TfmMain.sbtControlClick(Sender: TObject);
begin
 ShowLists;
end;

procedure TfmMain.fpMainExit(Sender: TObject);
begin
 PaintActive(False);
 FLastActive := Nil;
end;

procedure TfmMain.sbtOptionsClick(Sender: TObject);
begin
 if EditOptions then begin
  SetOptions;
  SaveOptions;
 end;
end;

procedure TfmMain.sbtZoomClick(Sender: TObject);
begin
 if Sender = sbtZoomIn then
  fpMain.Zoom(fpMain.Scale * 2, Nil)
 else
 if Sender = sbtZoomOut then
  fpMain.Zoom(fpMain.Scale div 2, Nil)
end;

procedure TfmMain.sbtAboutClick(Sender: TObject);
begin
 ShowAbout;
end;

procedure TfmMain.EventsChange(Sender: TObject);
begin
 if Assigned(fmControl) then fmControl.RefreshEvents;
end;

procedure TfmMain.SetDoorTimeOut(Door: TFlexDoor; PassId: integer);
var i: integer;
    TimeOut: PDoorTimeOut;
begin
 // Find door in persent timeouts
 TimeOut := Nil;
 for i:=0 to FDoorsTimeOuts.Count-1 do
  if PDoorTimeOut(FDoorsTimeOuts[i]).Door = Door then begin
   TimeOut := PDoorTimeOut(FDoorsTimeOuts[i]);
   break;
  end;
 if not Assigned(TimeOut) then begin
  // Create new timeout
  New(TimeOut);
  FillChar(TimeOut^, SizeOf(TimeOut^), 0);
  TimeOut.Door := Door;
  FDoorsTimeOuts.Add(TimeOut);
 end;
 TimeOut.PassId := PassId;
 TimeOut.TimeToClose := Now + SecToDT(3);
end;

function TfmMain.DoPass(PassId, PersonId: integer; DoOpen: boolean): boolean;
var Door: TFlexControl;
    Info: string;
    Index: integer;
    IsBlocked: boolean;
    NewDoorState: TDoorState;
    NewDoorEvent: TDoorEvent;
    NewAccessDir: TDoorAccessDirection;
begin
 Result := false;
 Door := fmMain.FindPass(PassId);
 if Assigned(Door) then with TFlexDoor(Door) do begin
  IsBlocked := DoorEvent = deBroken;
  // Define direction
  if PassAProp.Value = PassId
   then NewAccessDir := ddForward
   else NewAccessDir := ddBackward;
  // Define state
  if DoOpen then begin
   NewDoorState := dsOpened;
   if IsBlocked
    then Info := 'Attempt to open blocked door %d by '
    else Info := 'Door %d opened by ';
  end else begin
   NewDoorState := dsClosed;
   if IsBlocked
    then Info := 'Attempt to close blocked door %d by '
    else Info := 'Door %d closed by ';
  end;
  // Define person
  if PersonId = -1 then begin
   // is time out
   NewDoorEvent := DoorEvent; //deInactive;
   Info := Info + 'timeout';
  end else
  if PersonId = 0 then begin
   // is operator
   NewDoorEvent := deInactive;
   Info := Info + 'operator';
  end else begin
   // is person
   Index := dmMain.FindPerson(PersonId);
   if Index >= 0 then with PPerson(dmMain.Personnel[Index])^ do begin
    Info := Info + FirstName + ' ' + LastName;
    NewDoorEvent := deValid;
   end else begin
    // Alert. Unknown person
    Info := Info + '<unknown>';
    NewDoorEvent := deError;
   end;
  end;
  if not IsBlocked then begin
   // Set new state
   AccessDir := NewAccessDir;
   DoorEvent := NewDoorEvent;
   DoorState := NewDoorState;
   if DoorState = dsOpened then SetDoorTimeOut(TFlexDoor(Door), PassId);
   if not FMinimized {(WindowState <> wsMinimized)} and
      Options.ShowPersonsCards then
    if Options.ShowCardsInFlex then begin
     // Flex card
     if DoorState = dsOpened
      then ShowFlexCard(PassId, PersonId, Door)
      else DestroyFlexCard(Door);
    end else
    // Card as form
    if Door.ParentScheme = fpMain.ActiveScheme then
     if DoorState = dsOpened
      then ShowPersonCard(PassId, PersonId, Door)
      else HidePersonCard(Door);
   Result := true;
  end else begin
   // Alarm state
   Result := DoorState = dsOpened;
   if not Result then
    if Options.ShowCardsInFlex then
     // Flex card
     ShowFlexCard(PassId, PersonId, Door, True)
    else
    // Card as form
    if Door.ParentScheme = fpMain.ActiveScheme then
     ShowPersonCard(PassId, PersonId, Door, True);
  end;
  // Add event
  dmMain.AddEvent(Format(Info, [PassId]), PassId, 0, PersonId);
 end;
end;

procedure TfmMain.DoBlock(PassId: integer; IsBlock: boolean);
var Door: TFlexDoor;
    Info: string;
    IsBlocked: boolean;
begin
 Door := TFlexDoor(fmMain.FindPass(PassId));
 if not Assigned(Door) then exit;
 IsBlocked := Door.DoorEvent = deBroken;
 if IsBlocked = IsBlock then exit;
 if IsBlock then begin
  Info := 'Door %d blocked by operator';
  Door.DoorEvent := deBroken;
 end else begin
  Info := 'Door %d unblocked by operator';
  Door.DoorEvent := deInactive;
 end;
 // Add event
 dmMain.AddEvent(Format(Info, [PassId]), PassId, 0, 0);
end;

procedure TfmMain.tmTimeOutsTimer(Sender: TObject);
var i: integer;
begin
 for i:=FDoorsTimeOuts.Count-1 downto 0 do
  with PDoorTimeOut(FDoorsTimeOuts[i])^ do
   if Now - TimeToClose >= 0 then begin
    DoPass(PassId, -1, False);
    Dispose(PDoorTimeOut(FDoorsTimeOuts[i]));
    FDoorsTimeOuts.Delete(i);
   end;
end;

procedure TfmMain.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
 if (Key = VK_MENU) and (fpMain.ToolMode = ftmSelect) and
    not Assigned(fpMain.CreatingControlClass) then
  fpMain.ToolMode := ftmPan
 else
 if Key = vk_F1 then sbtAbout.Click;
end;

procedure TfmMain.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
 if (Key = VK_MENU) and
   ((fpMain.ToolMode = ftmPan) or (fpMain.ToolMode = ftmPanning)) then
  fpMain.ToolMode := ftmSelect;
end;

end.

